"use strict";

const fs = require('fs');
const path = require('path');

/**
 * Return the first valid element
 * 
 * @param {Array.<any>} array 
 */
function valueQueue(array) {
  for (var i = 0; i < array.length; i++) {
    if (typeof array[i] !== 'undefined') {
      return array[i];
    }
  }
}
exports.valueQueue = valueQueue;

/**
 * Round all number value in a javascript object at given decimal
 * 
 * @param {any} obj - Javascript object to be rounded 
 * @param {number} [decimal=0] - Round at this decimal 
 */
function roundAllValue (obj, decimal = 0) {
  Object.keys(obj).forEach(key => {
    if (typeof(obj[key]) === "object" && obj[key] !== null) {
      roundAllValue (obj[key], decimal);
    } else if(isNumeric(obj[key])) {
      const num = parseFloat(obj[key]);
      obj[key] = roundNumber(num, decimal);
    }
  });
}
exports.roundAllValue = roundAllValue;

/**
 * Round a given value at desire decimal
 * 
 * @param {number} num 
 * @param {number} scale 
 * @return {number}
 */
function roundNumber(num, scale) {
  if(!("" + num).includes("e")) {
    return +(Math.round(num + "e+" + scale)  + "e-" + scale);
  } else {
    const arr = ("" + num).split("e");
    let sig = ""
    if(+arr[1] + scale > 0) {
      sig = "+";
    }
    return +(Math.round(+arr[0] + "e" + sig + (+arr[1] + scale)) + "e-" + scale);
  }
}
exports.roundNumber = roundNumber;

/**
 * Stringify javascript object to BMFont compatible json or xml
 * 
 * @param {Object} data - Java object data 
 * @param {string} outputType - Type of output "xml"(default) "json"
 * 
 */
function stringify(data, outputType) {
  if (outputType === "json") {
    return toJSON(data);
  } else {
    return toBMFontXML(data);
  }
}
exports.stringify = stringify;

function toJSON(data) {
  return JSON.stringify(data);
}

/**
 * Stringify the given array, seperated by seperator
 * 
 * @param {any} array 
 * @param {string} [seperator=""] 
 * @returns {string} 
 */
function stringifyArray(array, seperator = "") {
  let result = "";
  let lastIndex = array.length - 1;
  array.forEach((element, index) => {
    result += element;
    if (index !== lastIndex){
      result += seperator;
    }
  });
  return result;
}
exports.stringifyArray = stringifyArray;

/**
 * Tell if the given object is string 
 * 
 * @param {any} n 
 * @returns {boolean} 
 */
function isString (s) {
  return (typeof s === 'string' || s instanceof String);
}
exports.isString = isString;

/**
 * Tell if the given object is numeric 
 * 
 * @param {any} n 
 * @returns {boolean} 
 */
function isNumeric (n) {
  return !isNaN(parseFloat(n)) && isFinite(n);
}
exports.isNumeric = isNumeric;

/**
 * Tell if the given object is empty
 * 
 * @param {any} obj 
 * @returns 
 */
function isEmpty (obj) {
  if (Object.getOwnPropertyNames(obj).length > 0) return false;
  else return true;
}
exports.isEmpty = isEmpty;

function insidePath (command, contours) {
  let x = command.x, y = command.y;
  let inside = false;
  contours.forEach(contour => {
    for (let i = 0, j = contour.length - 1; i < contour.length; j = i++) {
      let xi = contour[i].x, yi = contour[i].y;
      let xj = contour[j].x, yj = contour[j].y;

      let intersect = ((yi > y) != (yj > y)) && (x < (xj - xi) * (y - yi) / (yj - yi) + xi);
      if (intersect) inside = !inside;
    }
    if (inside) return;
  });
  return inside;
}

function isClockwise (contour) {
  let sum = 0;
  for (let i = 0; i < contour.length - 1; i++) {
      let command = contour[i], command_next = contour[i+1];
      if(command_next.type == 'Z') break;
      sum += (command_next.x - command.x) * (command_next.y + command.y);
  }
  return sum > 0
}

function reverseContour (contour) {
  let reversedContour = [];
  let tmpPoint = [];
  let hasCloseCmd = false;
  let isFirstCmd = true;

  for (let i = contour.length - 1; i > 0; i--) {
    let command = contour[i];
    let rev_comand = {};
    if (command.type === 'Z') {
      hasCloseCmd = true;
      continue;
    }
    if (isFirstCmd) {
      isFirstCmd = false;
      reversedContour.push({type: 'M', x: command.x, y: command.y});
    }
    rev_comand.type = command.type;
    rev_comand.x = contour[i - 1].x;
    rev_comand.y = contour[i - 1].y;
    if (command.type === 'C') {
      rev_comand.x1 = command.x2;
      rev_comand.y1 = command.y2;
      rev_comand.x2 = command.x1;
      rev_comand.y2 = command.y1;
    } else if (command.type === 'Q') {
      rev_comand.x1 = command.x1;
      rev_comand.y1 = command.y1;
    }
    reversedContour.push(rev_comand);
  }
  if (hasCloseCmd) {
    reversedContour.push({type: 'Z'});
  }
  return reversedContour;
}

/**
 * Align all contours' clockwiseness.
 * 
 * @param {Array} contours  Array of contour 
 * @param {boolean} direction  true for clockwise, false for counter-clockwise
 */
function alignClockwise(contours, direction) {
  let numReversed = 0;
  for (let i = 0; i < contours.length; i++) {
    let contour = contours[i];
    let restContours = contours.slice(0);
    restContours.splice(i, 1);
    if (contour.length === 0) continue;
    let isInside =  insidePath(contour[0], restContours) && 
                    insidePath(contour[Math.ceil(contour.length / 2)], restContours);
    let dir = isInside ? (!direction) : direction;
    if (isClockwise(contour) != dir) {
      contours[i] = reverseContour(contour);
      numReversed ++;
    }
  }
  return numReversed;
}
exports.alignClockwise = alignClockwise; 

/**
 * Convert contour commands to msdfgen shape description 
 * 
 * @param {Array} contours Array of font contour
 * @returns {string} 
 */
function stringifyContours(contours) {
  let shapeDesc = '';
  contours.forEach(contour => {
    shapeDesc += '{';
    const lastIndex = contour.length - 1;
    let _x, _y;
    contour.forEach((command, index) => {
      roundAllValue(command, 3);
      if (command.type === 'Z') {
        if(contour[0].x !== _x || contour[0].y !== _y) {
          shapeDesc += '# ';
        }
      } else {
        if (command.type === 'C') {
          shapeDesc += `(${command.x1}, ${command.y1}; ${command.x2}, ${command.y2}); `;
        } else if (command.type === 'Q') {
          shapeDesc += `(${command.x1}, ${command.y1}); `;
        }
        shapeDesc += `${command.x}, ${command.y}`;
        _x = command.x;
        _y = command.y;
        if (index !== lastIndex) {
          shapeDesc += '; ';
        }
      }
    });
    shapeDesc += '}';
  });
  return shapeDesc;
}
exports.stringifyContours = stringifyContours;

let pointTolerance = 0.5, areaTolerance = 2;
function setTolerance(pointValue, areaValue) {
  pointTolerance = pointValue;
  areaTolerance = areaValue;
}
exports.setTolerance = setTolerance;

function same(p1, p2) {
  return Math.abs(p1[0] - p2[0]) < pointTolerance && Math.abs(p1[1] - p2[1]) < pointTolerance;
}

class boundBox {
  constructor(left = 0, top = 0, right = 0, bottom = 0) {
    this.left = left;
    this.right = right;
    this.top = top;
    this.bottom = bottom;
    this.updateSize();
  }
 
  updateSize() {
    this.width = this.right - this.left;
    this.height = this.top - this.bottom;
  }

  update(x, y) {
    this.left = Math.min(this.left, x);
    this.right = Math.max(this.right, x);
    this.top = Math.max(this.top, y);
    this.bottom = Math.min(this.bottom, y);
    this.updateSize();
  }

  area() {
    this.updateSize();
    return this.width * this.height;
  }
}
module.exports.boundBox = boundBox;

function degenerate(p) {
  for (let i = 0; i < p.length - 1; i++) 
    if (same(p[i], p[i + 1])) p.splice(i, 1);
  return p.length;
} 

function isDegenerate(contour) {
  if (contour.length < 3) return true; // early quit with 0-area contour
  let bBox = new boundBox(contour[0].x, contour[0].y,contour[0].x, contour[0].y);
  for (let i = 0; i < contour.length - 1; i++) {
    let command_prev = contour[i];
    bBox.update(command_prev.x, command_prev.y);
    let command = contour[i + 1];
    let p = [[command_prev.x, command_prev.y]];
    if (command.type === 'C') { p.push([command.x1, command,y1]); p.push([command.x2, command.y2]); }
    else if (command.type === 'Q') p.push([command.x1, command.y1]);
    if (command.type === 'Z') p.push([contour[0].x, contour[0].y]);
    else p.push([command.x, command.y]); 
    
    let _nump = p.length;
    let nump = degenerate(p);
    if (nump < 2) {
      if (i === contour.length - 2 || command.type === 'Z') contour.splice(i, 1);
      else contour.splice(i + 1, 1);
      continue;
    } 
    // if point not degenerated, continue
    if (nump === _nump) continue;
    let newCommand = {};
    if (nump === 3) { 
      newCommand.type = 'Q';
      newCommand.x1 = p[1][0];
      newCommand.y1 = p[1][1];
      newCommand.x = p[2][0];
      newCommand.y = p[2][1];
    } else {
      newCommand.type = 'L';
      newCommand.x = p[1][0];
      newCommand.y = p[1][1];
    }
    contour[i + 1] = newCommand;
  }
  if (bBox.area() < areaTolerance || contour.length < 3) {
    return true;
  } else {
    return false;
  }
}

function filterContours(contours) {
  let filtered = 0;
  for (let i = 0; i < contours.length; i++) {
    if(isDegenerate(contours[i])) {
      contours.splice(i, 1);
      filtered ++;
    } 
  }
  return filtered;
}
exports.filterContours = filterContours;

/**
 * Return the largest length of string array.
 *
 * @param {Array.<String>} arr
 * @return {Number}
 * @api private
 */

function longestLength(arr) {
  return arr.reduce((max, element) => {
    return Math.max(max, element.length);
  }, 0);
}
exports.longestLength = longestLength;

/**
 * Pad `str` to `width`.
 *
 * @param {String} str
 * @param {Number} width
 * @return {String}
 * @api private
 */
function pad(str, width) {
  var len = Math.max(0, width - str.length);
  return str + Array(len + 1).join(' ');
}
exports.pad = pad;

/**
 * Find all files inside a dir, recursively.
 * @function getAllFiles
 * @param  {string} dir Dir path string.
 * @return {string[]} Array with all file names that are inside the directory.
 */
const getAllFiles = (dir, recursive = true) =>
  fs.readdirSync(dir).reduce((files, file) => {
    const name = path.join(dir, file);
    const isDirectory = fs.statSync(name).isDirectory();
    return isDirectory ? (recursive ? [...files, ...getAllFiles(name)] : [...files]) : [...files, name];
  }, []);
exports.getAllFiles = getAllFiles;


function getLeafFolder (pathalike) {
  const leafFolder = path.dirname(pathalike).split(path.sep).pop();
  return leafFolder ? leafFolder : "";
}
exports.getLeafFolder = getLeafFolder;