"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Atlasify = exports.Options = void 0;
const maxrects_packer_1 = require("maxrects-packer");
const jimp_1 = __importDefault(require("jimp"));
const path_1 = __importDefault(require("path"));
const pixelmatch_1 = __importDefault(require("pixelmatch"));
const sheet_1 = require("./geom/sheet");
const exporter_1 = require("./exporter");
const fs_1 = require("fs");
let appInfo = require('../package.json');
/**
 * Options class for atlasify and maxrects-packer
 *
 * @property {boolean} options.square use square size (default is true)
 * @property {boolean} options.pot use power of 2 sizing (default is true)
 * @property {boolean} options.square use square size (default is false)
 * @property {boolean} options.allowRotation allow rotation wihile packing (default is false)
 *
 * @export
 * @interface Option
 * @export
 * @class Options
 * @implements {IOption}
 */
class Options {
    /**
     * Creates an instance of Options.
     * @param {string} [name='sprite'] output filename of atlas/spreadsheet (default is 'sprite.png')
     * @param {number} [width=2048] ouput texture atlas width (defaut: 2048)
     * @param {number} [height=2048] ouput texture atlas height (defaut: 2048)
     * @param {number} [padding=0] padding between images (Default: 0)
     * @memberof Options
     */
    constructor(name = 'sprite', width = 2048, height = 2048, padding = 0, type = "JsonHash") {
        this.name = name;
        this.width = width;
        this.height = height;
        this.padding = padding;
        this.type = type;
        /**
         * Atlas will automaticly shrink to the smallest possible square
         *
         * @type {boolean}
         * @memberof Options
         */
        this.smart = true;
        /**
         * Atlas size shall be power of 2
         *
         * @type {boolean}
         * @memberof Options
         */
        this.pot = true;
        /**
         * Atlas size shall be square
         *
         * @type {boolean}
         * @memberof Options
         */
        this.square = false;
        /**
         * Allow 90-degree rotation while packing
         *
         * @type {boolean}
         * @memberof Options
         */
        this.allowRotation = false;
        /**
         * Controlling packer border to edge
         *
         * @type {number}
         * @memberof Options
         */
        this.border = 0;
        /**
         * Instant mode will skip sorting and pack using given array order
         *
         * @type {boolean}
         * @memberof Options
         */
        this.instant = false;
        /**
         * Seperate sheets packing based on folder
         *
         * @type {boolean}
         * @memberof Options
         */
        this.seperateFolder = false;
        /**
         * Group sheets packing based on folder
         *
         * @type {boolean}
         * @memberof Options
         */
        this.groupFolder = false;
        /**
         * Remove surrounding transparent pixels
         *
         * @type {boolean}
         * @memberof Options
         */
        this.trimAlpha = false;
        /**
         * Trim alpha with tolerence value
         *
         * @type {number}
         * @memberof Options
         */
        this.alphaTolerence = 0;
        /**
         * Extrude amount of edge pixels, will automaticly `trimAlpha` first.
         *
         * @type {number}
         * @memberof Options
         */
        this.extrude = 0;
        /**
         * Draw debug info onto atlas
         *
         * @type {boolean}
         * @memberof Options
         */
        this.debug = false;
        /**
         * Search duplicated dummy sprites to reduce atlas element
         *
         * @type {boolean}
         * @memberof Options
         */
        this.searchDummy = false;
    }
}
exports.Options = Options;
class Atlasify {
    /**
     * Creates an instance of Atlasify.
     *
     * @param {Options} options Atlasify Options class
     * @memberof Atlasify
     */
    constructor(options) {
        this.options = options;
        this._debugColor = 0xff000088;
        this._atlas = [];
        this._dirty = 0;
        this._spritesheets = [];
        this._inputPaths = [];
        this._sheets = [];
        if (options.seperateFolder)
            options.tag = true;
        if (options.groupFolder) {
            options.tag = true;
            options.exclusiveTag = false;
        }
        this._packer = new maxrects_packer_1.MaxRectsPacker(options.width, options.height, options.padding, options);
        this._exporter = new exporter_1.Exporter();
        this._exporter.setExportFormat(this.options.type);
    }
    /**
     * Add arrays of pathalike images url and do packing
     *
     * @param {string[]} paths
     * @param {(atlas: Atlas[], spritesheets: Spritesheet[]) => void} callback
     * @memberof Atlasify
     */
    addURLs(paths, callback) {
        this._inputPaths = this._inputPaths.concat(paths);
        let loader = paths.map((img) => __awaiter(this, void 0, void 0, function* () {
            return jimp_1.default.read(img).then(image => this.metricFromImage(image, img));
        }));
        return Promise.all(loader)
            .then(() => {
            return this.pack(callback);
        })
            .catch(err => {
            console.error("File load error : " + err);
            if (callback)
                callback(err);
            return Promise.reject(err);
        });
    }
    metricFromImage(image, imgPath) {
        const sheet = new sheet_1.Sheet(image.bitmap.width, image.bitmap.height);
        sheet.name = path_1.default.basename(imgPath);
        sheet.url = imgPath;
        sheet.data = image;
        // post-processing
        if (this.options.extrude > 0) {
            sheet.trimAlpha(this.options.alphaTolerence); // need to trim before extrude
            sheet.extrude(this.options.extrude);
        }
        else if (this.options.trimAlpha) {
            sheet.trimAlpha(this.options.alphaTolerence);
        }
        if (this.options.seperateFolder || this.options.groupFolder) {
            const tag = this.getLeafFolder(imgPath);
            if (tag)
                sheet.tag = tag;
        }
        // search if image already exist
        let isNew = true;
        for (const i in this._sheets) {
            const s = this._sheets[i];
            if (!s.data)
                continue; // skip empty sheet
            if (s.name === sheet.name) {
                isNew = false;
                if (s.width === sheet.width &&
                    s.height === sheet.height &&
                    this.options.searchDummy && // do pHash compare only on same size image
                    s.hash === sheet.hash) {
                    // deep pixel compare
                    let diff = pixelmatch_1.default(s.data.bitmap.data, sheet.data.bitmap.data, null, s.width, s.height);
                    if (diff === 0)
                        return; // early exit if no change
                }
                // input image has changed, need process
                this._sheets[i] = sheet;
                break;
            }
            else if (this.options.searchDummy && s.width === sheet.width &&
                s.height === sheet.height &&
                s.hash === sheet.hash) {
                // deep pixel compare
                let diff = pixelmatch_1.default(s.data.bitmap.data, sheet.data.bitmap.data, null, s.width, s.height);
                if (diff !== 0)
                    break; // different image
                if (s.dummy.includes(sheet.name))
                    return; // already in dummy list, early exit
                // This is the dummy sheet with different name
                isNew = false;
                s.dummy.push(sheet.name);
                break;
            }
        }
        // if no early exit, set dirty status
        this._dirty++;
        // push to _sheets if is new sheet
        if (isNew) {
            this._sheets.push(sheet);
            if (this.options.instant) {
                this._packer.add(sheet);
            }
        }
    }
    pack(callback) {
        if (this._dirty === 0)
            return Promise.resolve(this); // early quick if nothing changed
        let ext = path_1.default.extname(this.options.name);
        const basename = path_1.default.basename(this.options.name, ext);
        if (ext === "")
            ext = "png"; // assign default format PNG
        else
            ext = ext.slice(1).toLowerCase(); // trim . of extname
        const fillColor = (ext === "png") ? 0x00000000 : 0x000000ff;
        const tagCount = {};
        if (!this.options.instant) {
            this._packer.reset();
            this._packer.addArray(this._sheets);
        }
        else {
            this._packer.repack(false);
        }
        this._packer.bins.forEach((bin, index) => {
            // Count tags
            let tag = bin.tag ? bin.tag : "_";
            if (!tagCount.hasOwnProperty(tag))
                tagCount[tag] = 0; // create index key if not exist
            else
                tagCount[tag]++;
            if (!bin.dirty)
                return; // early return if bin is not changed
            let binName = basename;
            if (bin.tag)
                binName = `${bin.tag}-${binName}`;
            this._atlas[index] = {
                id: tagCount[tag],
                width: bin.width,
                height: bin.height,
                image: new jimp_1.default(bin.width, bin.height, fillColor),
                name: binName,
                format: "RGBA8888",
                ext: ext
            };
            if (bin.tag)
                this._atlas[index].tag = bin.tag;
            const image = this._atlas[index].image;
            const serializedSheet = [];
            // Render rects onto atlas
            bin.rects.forEach(rect => {
                const sheet = rect;
                const buffer = sheet.data;
                // sheet.frame.x += sheet.x;
                // sheet.frame.y += sheet.y;
                if (this.options.debug) {
                    const debugFrame = new jimp_1.default(sheet.frame.width, sheet.frame.height, this._debugColor);
                    image.blit(debugFrame, sheet.frame.x, sheet.frame.y);
                }
                image.composite(buffer, sheet.x, sheet.y);
                // share rects iteration to add serialized sheets
                serializedSheet.push(rect.serialize());
                if (rect.dummy.length > 0) {
                    rect.dummy.forEach(n => {
                        serializedSheet.push(Object.assign(Object.assign({}, rect.serialize()), { name: n }));
                    });
                }
            });
            // Add tag to the last sheet to control mustache trailing comma
            serializedSheet[serializedSheet.length - 1].last = true;
            // prepare spritesheet data
            this._spritesheets[index] = {
                id: tagCount[tag],
                name: binName,
                imageName: `${binName}.${ext}`,
                imageFormat: "RGBA8888",
                width: bin.width,
                height: bin.height,
                scale: 1,
                rects: serializedSheet,
                format: this.options.type,
                ext: this._exporter.getExtension(),
                appInfo: appInfo
            };
            if (bin.tag)
                this._spritesheets[index].tag = bin.tag;
        });
        // remove id if tag count < 2
        this.pruneTagIndex(tagCount);
        this._dirty = 0; // set clean
        if (callback)
            callback(undefined, this._atlas, this._spritesheets);
        return Promise.resolve(this);
    }
    addBuffers(buffers, callback) {
        // TODO
    }
    /**
     * Enclose previous packing bin and start a new one.
     *
     * @returns {number}
     * @memberof Atlasify
     */
    next() {
        this._packer.next();
        return this._packer.currentBinIndex;
    }
    /**
     * Asycn save current project & settings to file and return serialized string
     *
     * @param {boolean} [humanReadable=false]
     * @param {string} [pathalike]
     * @returns {Promise<string}>}
     * @memberof Atlasify
     */
    save(...args) {
        return __awaiter(this, void 0, void 0, function* () {
            const atlasBase64 = yield Promise.all(this._atlas.map((a) => __awaiter(this, void 0, void 0, function* () { return a.image.getBase64Async(jimp_1.default.MIME_PNG); })));
            const atl = {
                options: this.options,
                packer: this._packer.save(),
                spritesheets: this._spritesheets,
                atlas: this._atlas.map((a, i) => {
                    return {
                        id: a.id ? a.id : 0,
                        width: a.width,
                        height: a.height,
                        name: a.name,
                        format: "RGBA8888",
                        ext: a.ext,
                        image: atlasBase64[i]
                    };
                }),
                imagePaths: this._inputPaths
            };
            let humanReadable = false;
            let pathalike;
            if (args.length === 0) {
                humanReadable = false;
            }
            else if (args.length === 1) {
                if (typeof (args[0]) === "boolean") {
                    humanReadable = args[0];
                }
                else if (typeof (args[0] === "string")) {
                    pathalike = args[0];
                }
                else {
                    throw new Error("Atlasify.save(): wrong argument type");
                }
            }
            else if (args.length > 1) {
                if (typeof (args[0]) === "boolean" && typeof (args[1]) === "string") {
                    humanReadable = args[0];
                    pathalike = args[1];
                }
                else {
                    throw new Error("Atlasify.save(): wrong argument type");
                }
            }
            const result = humanReadable ? JSON.stringify(atl, null, 2) : JSON.stringify(atl);
            if (pathalike) {
                fs_1.writeFile(pathalike, result, err => {
                    if (err) {
                        console.error(`Saving atl file encountered error: ${err}`);
                        return false;
                    }
                    else {
                        console.log(`Saved configuration: ${pathalike}`);
                        return true;
                    }
                });
            }
            return result;
        });
    }
    static Load(pathalike, overrides = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const factory = new Atlasify(new Options());
            return factory.load(pathalike, overrides);
        });
    }
    load(pathalike, overrides = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const atl = JSON.parse(fs_1.readFileSync(pathalike, 'utf-8'));
            this._sheets = [];
            this.options = Object.assign(Object.assign({}, atl.options), overrides); // combining saved options and cli options
            this._packer = new maxrects_packer_1.MaxRectsPacker(this.options.width, this.options.height, this.options.padding, this.options);
            this._exporter = new exporter_1.Exporter();
            this._exporter.setExportFormat(this.options.type);
            // Load packer
            this._packer.load(atl.packer);
            // Load spritesheets
            this._spritesheets = atl.spritesheets;
            // load atlas
            this._atlas = yield Promise.all(atl.atlas.map((a, i) => __awaiter(this, void 0, void 0, function* () {
                // async overwrite atlas base64 string image with Jimp object
                return Object.assign(Object.assign({}, a), { image: yield jimp_1.default.read(Buffer.from(a.image.replace(/^data:image\/png;base64,/, ""), 'base64')) });
            })));
            // Load sheets
            const loaders = [];
            this._spritesheets.forEach((spritesheet, i) => {
                spritesheet.rects.forEach(r => {
                    const sheet = sheet_1.Sheet.Factory(r);
                    const loader = jimp_1.default.read(sheet.url);
                    loader
                        .then(image => {
                        const reloaded = new sheet_1.Sheet(image.bitmap.width, image.bitmap.height);
                        reloaded.data = image;
                        // post-processing
                        // Note: only use saved options which sync saved sheet metrics
                        // option overriding will be done in deep repack using inputPath
                        if (atl.options.extrude > 0) {
                            reloaded.trimAlpha(atl.options.alphaTolerence); // need to trim before extrude
                            reloaded.extrude(atl.options.extrude);
                        }
                        else if (atl.options.trimAlpha) {
                            reloaded.trimAlpha(atl.options.alphaTolerence);
                        }
                        // unrotate sheets for stable result
                        if (sheet.rot)
                            sheet.rot = false;
                        sheet.data = reloaded.data;
                        // manage option overrides
                        if (!this.options.seperateFolder && sheet.tag)
                            delete sheet.tag;
                    })
                        .catch(error => {
                        console.error(error);
                        console.log("Fall back to pre-rendered atlas");
                        sheet.data = new jimp_1.default(sheet.width, sheet.height)
                            .blit(this._atlas[i].image, 0, 0, sheet.x, sheet.y, sheet.width, sheet.height);
                        // unrotate sheets for stable result
                        if (sheet.rot)
                            sheet.rot = false;
                    });
                    this._sheets.push(sheet);
                    this._packer.bins[i].rects.push(sheet);
                    loaders.push(loader);
                });
            });
            // TODO: Load imagePaths
            yield Promise.all(loaders);
            console.log("Load completed");
            return this;
        });
    }
    /**
     * Get all atlas/image array
     *
     * note: this will only available with all async image load & packing done.
     *
     * @readonly
     * @type {Atlas[]}
     * @memberof Atlasify
     */
    get atlas() { return this._atlas; }
    /**
     * Get all serialized spritesheets array.
     *
     * note: this will only available with all async image load & packing done.
     *
     * @readonly
     * @type {Spritesheet[]}
     * @memberof Atlasify
     */
    get spritesheets() { return this._spritesheets; }
    get exporter() { return this._exporter; }
    getLeafFolder(pathalike) {
        const leafFolder = path_1.default.dirname(pathalike).split(path_1.default.sep).pop();
        return leafFolder;
    }
    pruneTagIndex(tagCount) {
        for (let a of this._atlas) {
            const tag = a.tag ? a.tag : "_";
            if (tagCount[tag] < 1 && a.hasOwnProperty("id"))
                delete a.id;
            if (a.hasOwnProperty("id")) {
                a.name = `${a.name}.${a.id}`; // append index to image filename
            }
        }
        for (let s of this._spritesheets) {
            const tag = s.tag ? s.tag : "_";
            if (tagCount[tag] < 1 && s.hasOwnProperty("id"))
                delete s.id;
            if (s.hasOwnProperty("id")) {
                s.name = `${s.name}.${s.id}`; // append index to spritesheet filename
                const ext = path_1.default.extname(s.imageName);
                s.imageName = `${s.name}${ext}`;
            }
        }
    }
}
exports.Atlasify = Atlasify;
var sheet_2 = require("./geom/sheet");
Object.defineProperty(exports, "Sheet", { enumerable: true, get: function () { return sheet_2.Sheet; } });
//# sourceMappingURL=atlasify.js.map