/// <reference types="node" />
import { IOption, IBin } from "maxrects-packer";
import Jimp from "jimp";
import { Exporter } from "./exporter";
/**
 * Options class for atlasify and maxrects-packer
 *
 * @property {boolean} options.square use square size (default is true)
 * @property {boolean} options.pot use power of 2 sizing (default is true)
 * @property {boolean} options.square use square size (default is false)
 * @property {boolean} options.allowRotation allow rotation wihile packing (default is false)
 *
 * @export
 * @interface Option
 * @export
 * @class Options
 * @implements {IOption}
 */
export declare class Options implements IOption {
    name: string;
    width: number;
    height: number;
    padding: number;
    type: string;
    /**
     * Atlas will automaticly shrink to the smallest possible square
     *
     * @type {boolean}
     * @memberof Options
     */
    smart: boolean;
    /**
     * Atlas size shall be power of 2
     *
     * @type {boolean}
     * @memberof Options
     */
    pot: boolean;
    /**
     * Atlas size shall be square
     *
     * @type {boolean}
     * @memberof Options
     */
    square: boolean;
    /**
     * Allow 90-degree rotation while packing
     *
     * @type {boolean}
     * @memberof Options
     */
    allowRotation: boolean;
    /**
     * Controlling packer border to edge
     *
     * @type {number}
     * @memberof Options
     */
    border: number;
    /**
     * Instant mode will skip sorting and pack using given array order
     *
     * @type {boolean}
     * @memberof Options
     */
    instant: boolean;
    /**
     * Seperate sheets packing based on folder
     *
     * @type {boolean}
     * @memberof Options
     */
    seperateFolder: boolean;
    tag?: boolean;
    /**
     * Group sheets packing based on folder
     *
     * @type {boolean}
     * @memberof Options
     */
    groupFolder: boolean;
    exclusiveTag?: boolean;
    /**
     * Remove surrounding transparent pixels
     *
     * @type {boolean}
     * @memberof Options
     */
    trimAlpha: boolean;
    /**
     * Trim alpha with tolerence value
     *
     * @type {number}
     * @memberof Options
     */
    alphaTolerence: number;
    /**
     * Extrude amount of edge pixels, will automaticly `trimAlpha` first.
     *
     * @type {number}
     * @memberof Options
     */
    extrude: number;
    /**
     * Draw debug info onto atlas
     *
     * @type {boolean}
     * @memberof Options
     */
    debug: boolean;
    /**
     * Search duplicated dummy sprites to reduce atlas element
     *
     * @type {boolean}
     * @memberof Options
     */
    searchDummy: boolean;
    /**
     * Creates an instance of Options.
     * @param {string} [name='sprite'] output filename of atlas/spreadsheet (default is 'sprite.png')
     * @param {number} [width=2048] ouput texture atlas width (defaut: 2048)
     * @param {number} [height=2048] ouput texture atlas height (defaut: 2048)
     * @param {number} [padding=0] padding between images (Default: 0)
     * @memberof Options
     */
    constructor(name?: string, width?: number, height?: number, padding?: number, type?: string);
}
export declare type Atlas = {
    image: Jimp;
    ext: string;
    width: number;
    height: number;
    name: string;
    id?: number;
    tag?: string;
    format?: string;
};
export declare type Spritesheet = {
    name: string;
    id?: number;
    tag?: string;
    imageName: string;
    imageFormat: string;
    width: number;
    height: number;
    format: string;
    ext: string;
    scale: number;
    rects: object[];
    appInfo?: any;
    base64Data?: Base64Data;
};
export declare type Base64Data = {
    prefix: string;
    data: string;
};
export interface IAtl {
    options: Options;
    packer: IBin[];
    spritesheets: Spritesheet[];
    atlas: {
        image: string;
        ext: string;
        width: number;
        height: number;
        name: string;
        id?: number;
        tag?: string;
        format?: string;
    }[];
    imagePaths: string[];
}
export declare class Atlasify {
    options: Options;
    /**
     * Creates an instance of Atlasify.
     *
     * @param {Options} options Atlasify Options class
     * @memberof Atlasify
     */
    constructor(options: Options);
    /**
     * Add arrays of pathalike images url and do packing
     *
     * @param {string[]} paths
     * @param {(atlas: Atlas[], spritesheets: Spritesheet[]) => void} callback
     * @memberof Atlasify
     */
    addURLs(paths: string[], callback?: (err?: Error, atlas?: Atlas[], spritesheets?: Spritesheet[]) => void): Promise<Atlasify>;
    private metricFromImage;
    pack(callback?: ((err?: Error, atlas?: Atlas[], spritesheets?: Spritesheet[]) => void)): Promise<this>;
    addBuffers(buffers: Buffer[], callback: (atlas: Atlas[], spritesheets: Spritesheet[]) => void): void;
    /**
     * Enclose previous packing bin and start a new one.
     *
     * @returns {number}
     * @memberof Atlasify
     */
    next(): number;
    /**
     * Async serialize current project & settings to string
     *
     * @param {boolean} [humanReadable=false]
     * @returns {Promise<string>}
     * @memberof Atlasify
     */
    save(humanReadable?: boolean): Promise<string>;
    /**
     * Asycn save current project & settings to file
     *
     * @param {boolean} [humanReadable=false]
     * @param {string} [pathalike]
     * @returns {Promise<boolean>}
     * @memberof Atlasify
     */
    save(humanReadable?: boolean, pathalike?: string): Promise<boolean>;
    static Load(pathalike: string, overrides?: any): Promise<Atlasify>;
    load(pathalike: string, overrides?: any): Promise<Atlasify>;
    private _inputPaths;
    private _sheets;
    private _packer;
    private _debugColor;
    private _atlas;
    private _dirty;
    /**
     * Get all atlas/image array
     *
     * note: this will only available with all async image load & packing done.
     *
     * @readonly
     * @type {Atlas[]}
     * @memberof Atlasify
     */
    get atlas(): Atlas[];
    private _spritesheets;
    /**
     * Get all serialized spritesheets array.
     *
     * note: this will only available with all async image load & packing done.
     *
     * @readonly
     * @type {Spritesheet[]}
     * @memberof Atlasify
     */
    get spritesheets(): Spritesheet[];
    private _exporter;
    get exporter(): Exporter;
    private getLeafFolder;
    private pruneTagIndex;
}
export { Sheet } from './geom/sheet';
